"""GPIO pwm driver."""
import pigpio

from pwmled.driver import Driver


class GpioDriver(Driver):
    """Represents a pwm driver, which uses GPIOs of a Raspberry Pi."""

    RESOLUTION = 8

    def __init__(self, pins, freq=200, host='localhost', port=8888):
        """
        Initialize the driver.

        :param pins: The pin numbers, that should be controlled.
        :param freq: The pwm frequency.
        :param host: The host name of the Pi on which pigpio is running.
        :param port: The port on which pigpio is running.
        """
        super().__init__(pins, self.RESOLUTION, freq)

        self._pi = pigpio.pi(host, port, show_errors=False)
        if not self._pi.connected:
            raise ConnectionError('Could not connect to the pigpio daemon')

        for pin in self._pins:
            self._pi.set_PWM_frequency(pin, freq)

    def _set_pwm(self, raw_values):
        """
        Set pwm values on the controlled pins.

        :param raw_values: Raw values to set (0-255).
        """
        for pin, value in zip(self._pins, raw_values):
            try:
                current_value = self._pi.get_PWM_dutycycle(pin)
            except pigpio.error as error:
                if error.value == 'GPIO is not in use for PWM':
                    current_value = 0
                else:
                    raise

            if current_value != value:
                self._pi.set_PWM_dutycycle(pin, value)

    def _stop(self):
        """Stop the driver and release resources."""
        self._pi.stop()



"""Transition of a led."""
import time
import threading

from pwmled import Color


class Transition:
    """Represents a transition of a led."""

    def __init__(self, led, duration, src_state, dest_state):
        """
        Initialize the transition.

        :param led: The led to control.
        :param duration: The duration.
        :param src_state: The source state of the led.
        :param dest_state: The target state of the led.
        """
        self._led = led
        self._duration = duration
        self._src_state = src_state
        self._dest_state = dest_state

        self._cancelled = False
        self._finish_event = threading.Event()
        self._start_time = time.perf_counter()

    @property
    def duration(self):
        """
        Duration property.

        :return: The duration of the transition.
        """
        return self._duration

    @property
    def progress(self):
        """
        Progress property.

        :return: The progress of the transition (0.0-1.0).
        """
        if self._duration == 0:
            return 1

        run_time = time.perf_counter() - self._start_time
        return max(0, min(1, run_time / self._duration))

    @property
    def finished(self):
        """
        Finshed property.

        :return: True, if transition has finished. False otherwise.
        """
        return self._finish_event.is_set()

    @property
    def cancelled(self):
        """
        Cancelled property.

        :return: True, if transition was cancelled. False otherwise.
        """
        return self._cancelled

    def step(self):
        """Apply the current stage of the transition based on current time."""
        if self.cancelled or self.finished:
            return

        if self.progress == 1:
            self._finish()
            return

        state = {}
        src_is_on = self._src_state.get('is_on')
        dest_is_on = self._dest_state.get('is_on')
        if dest_is_on:
            state['is_on'] = True

        src_brightness = self._src_state.get('brightness')
        dest_brightness = self._dest_state.get('brightness')
        if src_is_on is False:
            if dest_brightness is None:
                dest_brightness = src_brightness
            src_brightness = 0
        if dest_is_on is False:
            dest_brightness = 0
        if src_brightness is not None and dest_brightness is not None:
            state['brightness'] = self._interpolate(
                src_brightness,
                dest_brightness,
            )

        src_color = self._src_state.get('color')
        dest_color = self._dest_state.get('color')
        if src_is_on is False:
            src_color = dest_color
        if src_color is not None and dest_color is not None:
            state['color'] = Color(*(
                self._interpolate(src_color[i], dest_color[i])
                for i in range(3)
            ))

        self._led.set(**state, cancel_transition=False)

    def _interpolate(self, start, end):
        """
        Interpolate a value from start to end at the current progress.

        :param start: The start value.
        :param end: The end value.
        :return: The interpolated value at the current progress.
        """
        diff = end - start
        return start + self.progress * diff

    def _finish(self):
        """Complete transition and mark it as finished."""
        state = self._dest_state.copy()
        if state.get('is_on') is False and state.get('brightness') is None:
            # If led was turned off, set brightness to initial value
            # so that the brightness is restored when it is turned on again
            state['brightness'] = self._src_state.get('brightness')
        self._led.set(**state, cancel_transition=False)

        self._finish_event.set()

    def wait(self, timeout=None):
        """
        Wait for transition to be finished.

        :param timeout: Timeout of the operation in seconds.
        """
        self._finish_event.wait(timeout=timeout)

    def cancel(self):
        """Cancel the transition."""
        self._cancelled = True
        self._finish_event.set()



"""Manager for led transitions."""
import time
import threading
from singleton import Singleton


class TransitionManager(object, metaclass=Singleton):
    """Represents a manager that executes transitions in a separate thread."""

    STEP_TIME = 0.001

    def __init__(self):
        """Initialize the manager."""
        self._thread = None
        self._transitions = []

    def execute(self, transition):
        """
        Queue a transition for execution.

        :param transition: The transition
        :return: The started transition.
        """
        self._transitions.append(transition)
        if self._thread is None or not self._thread.is_alive():
            self._thread = threading.Thread(
                target=self._transition_loop,
                daemon=True,
            )
            self._thread.start()

        return transition

    def _transition_loop(self):
        """Execute all queued transitions step by step."""
        while self._transitions:
            for transition in self._transitions:
                transition.step()
                if transition.finished:
                    self._transitions.remove(transition)

            time.sleep(self.STEP_TIME)



"""Simple led controller."""

from pwmled.transitions.transition import Transition
from pwmled.transitions.transition_manager import TransitionManager


class SimpleLed:
    """Represents a simple, one-color led that can be controlled."""

    def __init__(self, driver):
        """
        Initialize the led.

        :param driver: The driver that is used to control the led.
        """
        self._driver = driver
        self._is_on = False
        self._brightness = 1
        self._active_transition = None

    @property
    def is_on(self):
        """
        On-off state property.

        :return: True, if led is on. False otherwise.
        """
        return self._is_on

    def on(self):
        """Turn the led on."""
        self.set(is_on=True)

    def off(self):
        """Turn the led off."""
        self.set(is_on=False)

    @property
    def brightness(self):
        """
        Brightness property.

        :return: The brightness of the led (0.0-1.0).
        """
        return self._brightness

    @brightness.setter
    def brightness(self, brightness):
        """
        Set the brightness of the led.

        :param brightness: The brightness to set (0.0-1.0).
        """
        self.set(brightness=brightness)

    @property
    def state(self):
        """
        State property.

        :return: The current state of the led.
        """
        return dict(
            is_on=self.is_on,
            brightness=self.brightness,
        )

    def set(self, is_on=None, brightness=None, cancel_transition=True):
        """
        Set properties of the led simultaneously before updating pwm values.

        :param is_on: On-off state of the led.
        :param brightness: Brightness of the led.
        :param cancel_transition: Cancel active transitions.
        """
        if cancel_transition:
            self._cancel_active_transition()

        if is_on is not None:
            self._is_on = is_on

        if brightness is not None:
            self._assert_is_brightness(brightness)
            self._brightness = brightness

        self._update_pwm()

    def _update_pwm(self):
        """Update the pwm values of the driver regarding the current state."""
        if self._is_on:
            values = self._get_pwm_values()
        else:
            values = [0] * len(self._driver.pins)

        self._driver.set_pwm(values)

    def _get_pwm_values(self, brightness=None):
        """
        Get the pwm values for a specific state of the led.

        If a state argument is omitted, current value is used.

        :param brightness: The brightness of the state.
        :return: The pwm values.
        """
        if brightness is None:
            brightness = self.brightness

        return [brightness]

    def transition(self, duration, **dest_state):
        """
        Transition to the specified state of the led.

        If another transition is already running, it is aborted.

        :param duration: The duration of the transition.
        :param dest_state: The state to transition to.
        """
        self._assert_is_valid_state(dest_state)

        self._cancel_active_transition()
        return TransitionManager().execute(Transition(
            self,
            duration,
            self.state,
            dest_state,
        ))

    def _cancel_active_transition(self):
        if self._active_transition:
            self._active_transition.cancel()
            self._active_transition = None

    @classmethod
    def _assert_is_valid_state(cls, value):
        """
        Assert that the given value is a valid state object.

        :param value: The value to check.
        """
        brightness = value.get('brightness')
        if brightness is not None:
            cls._assert_is_brightness(brightness)

    @staticmethod
    def _assert_is_brightness(value):
        """
        Assert that the given value is a valid brightness.

        :param value: The value to check.
        """
        if not 0 <= value <= 1:
            raise ValueError('Brightness must be between 0 and 1.')



"""PCA9685 pwm driver."""
import busio
import adafruit_pca9685
from adafruit_blinka.agnostic import board_id
try:
    import board
except NotImplementedError:
    # Defer raising error to initialization of Pca9685Driver class
    board = None

from pwmled.driver import Driver


class Pca9685Driver(Driver):
    """Represents a pwm driver, which uses the pins of an PCA9685 I2C board."""

    RESOLUTION = 16

    def __init__(self, pins, freq=200, address=0x40):
        """
        Initialize the driver.

        :param pins: The pin numbers, that should be controlled.
        :param freq: The pwm frequency.
        :param address: The address of the PCA9685.
        """
        super().__init__(pins, self.RESOLUTION, freq)

        # Raise error if board couldn't be imported
        if not board:
            if not board_id:
                raise NotImplementedError('Board is unknown')
            raise NotImplementedError(f'Board "{board_id}" is not supported')

        i2c = busio.I2C(board.SCL, board.SDA)
        self._device = adafruit_pca9685.PCA9685(i2c, address=address)
        self._device.frequency = self._freq

    def _set_pwm(self, raw_values):
        """
        Set pwm values on the controlled pins.

        :param raw_values: Raw values to set (0-65535).
        """
        for pin, value in zip(self._pins, raw_values):
            channel = self._device.channels[pin]
            if channel.duty_cycle != value:
                channel.duty_cycle = value

    def _stop(self):
        """Stop the driver and release resources."""
        self._device.deinit()
